import argparse
import os.path as osp
import torch
import torch.nn.functional as F 
from torch_geometric.datasets import TUDataset
from torch_geometric.loader import DataLoader
from torch_geometric.nn import MLP, GINConv

class GIN(torch.nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels, num_layers=5):
        super().__init__()
        self.convs = torch.nn.ModuleList()
        for _ in range(num_layers):
            mlp = MLP([in_channels, hidden_channels, hidden_channels])
            self.convs.append(GINConv(nn=mlp, train_eps=False))
            in_channels = hidden_channels
        self.mlp = MLP([hidden_channels, hidden_channels, out_channels],batch_norm=False, dropout=0.5)
 
    def forward(self, x, edge_index):
        for conv in self.convs:
            x = conv(x, edge_index)
        return self.mlp(x)
 
 

